@@ -3,7 +3,6 @@ package main
 import (
 	"database/sql"
 	"errors"
-	"flag"
 	"fmt"
 	"io/ioutil"
 	"math"
@@ -11,39 +10,32 @@ import (
 	"net/url"
 	"os"
 	"regexp"
+	"runtime"
 	"strconv"
+	"strings"
 	"sync"
 	"time"
 
+	"gopkg.in/alecthomas/kingpin.v2"
 	"gopkg.in/yaml.v2"
 
+	"crypto/sha256"
 	"github.com/blang/semver"
 	_ "github.com/lib/pq"
 	"github.com/prometheus/client_golang/prometheus"
+	"github.com/prometheus/client_golang/prometheus/promhttp"
 	"github.com/prometheus/common/log"
 )
 
-var Version string = "0.0.1"
-
-var db *sql.DB = nil
+// Version is set during build to the git describe version
+// (semantic version)-(commitish) form.
+var Version = "0.0.1"
 
 var (
-	listenAddress = flag.String(
-		"web.listen-address", ":9187",
-		"Address to listen on for web interface and telemetry.",
-	)
-	metricPath = flag.String(
-		"web.telemetry-path", "/metrics",
-		"Path under which to expose metrics.",
-	)
-	queriesPath = flag.String(
-		"extend.query-path", "",
-		"Path to custom queries to run.",
-	)
-	onlyDumpMaps = flag.Bool(
-		"dumpmaps", false,
-		"Do not run, simply dump the maps.",
-	)
+	listenAddress = kingpin.Flag("web.listen-address", "Address to listen on for web interface and telemetry.").Default(":9187").String()
+	metricPath    = kingpin.Flag("web.telemetry-path", "Path under which to expose metrics.").Default("/metrics").String()
+	queriesPath   = kingpin.Flag("extend.query-path", "Path to custom queries to run.").Default("").String()
+	onlyDumpMaps  = kingpin.Flag("dumpmaps", "Do not run, simply dump the maps.").Bool()
 )
 
 // Metric name parts.
@@ -57,21 +49,22 @@ const (
 	staticLabelName = "static"
 )
 
-// landingPage contains the HTML served at '/'.
-// TODO: Make this nicer and more informative.
-var landingPage = []byte(`<html>
-<head><title>Postgres exporter</title></head>
-<body>
-<h1>Postgres exporter</h1>
-<p><a href='` + *metricPath + `'>Metrics</a></p>
-</body>
-</html>
-`)
-
+// ColumnUsage should be one of several enum values which describe how a
+// queried row is to be converted to a Prometheus metric.
 type ColumnUsage int
 
-// Implements the yaml.Unmarshaller interface
-func (this *ColumnUsage) UnmarshalYAML(unmarshal func(interface{}) error) error {
+// nolint: golint
+const (
+	DISCARD      ColumnUsage = iota // Ignore this column
+	LABEL        ColumnUsage = iota // Use this column as a label
+	COUNTER      ColumnUsage = iota // Use this column as a counter
+	GAUGE        ColumnUsage = iota // Use this column as a gauge
+	MAPPEDMETRIC ColumnUsage = iota // Use this column with the supplied mapping of text values
+	DURATION     ColumnUsage = iota // This column should be interpreted as a text duration (and converted to milliseconds)
+)
+
+// UnmarshalYAML implements the yaml.Unmarshaller interface.
+func (cu *ColumnUsage) UnmarshalYAML(unmarshal func(interface{}) error) error {
 	var value string
 	if err := unmarshal(&value); err != nil {
 		return err
@@ -82,21 +75,12 @@ func (this *ColumnUsage) UnmarshalYAML(unmarshal func(interface{}) error) error
 		return err
 	}
 
-	*this = columnUsage
+	*cu = columnUsage
 	return nil
 }
 
-const (
-	DISCARD      ColumnUsage = iota // Ignore this column
-	LABEL        ColumnUsage = iota // Use this column as a label
-	COUNTER      ColumnUsage = iota // Use this column as a counter
-	GAUGE        ColumnUsage = iota // Use this column as a gauge
-	MAPPEDMETRIC ColumnUsage = iota // Use this column with the supplied mapping of text values
-	DURATION     ColumnUsage = iota // This column should be interpreted as a text duration (and converted to milliseconds)
-)
-
 // Regex used to get the "short-version" from the postgres version field.
-var versionRegex = regexp.MustCompile(`^\w+ (\d+\.\d+\.\d+)`)
+var versionRegex = regexp.MustCompile(`^\w+ ((\d+)(\.\d+)?(\.\d+)?)\s`)
 var lowestSupportedVersion = semver.MustParse("9.1.0")
 
 // Parses the version of postgres into the short version string we can use to
@@ -104,13 +88,13 @@ var lowestSupportedVersion = semver.MustParse("9.1.0")
 func parseVersion(versionString string) (semver.Version, error) {
 	submatches := versionRegex.FindStringSubmatch(versionString)
 	if len(submatches) > 1 {
-		return semver.Make(submatches[1])
+		return semver.ParseTolerant(submatches[1])
 	}
 	return semver.Version{},
 		errors.New(fmt.Sprintln("Could not find a postgres version in string:", versionString))
 }
 
-// User-friendly representation of a prometheus descriptor map
+// ColumnMapping is the user-friendly representation of a prometheus descriptor map
 type ColumnMapping struct {
 	usage             ColumnUsage        `yaml:"usage"`
 	description       string             `yaml:"description"`
@@ -118,22 +102,20 @@ type ColumnMapping struct {
 	supportedVersions semver.Range       `yaml:"pg_version"`     // Semantic version ranges which are supported. Unsupported columns are not queried (internally converted to DISCARD).
 }
 
-func (this *ColumnMapping) UnmarshalYAML(unmarshal func(interface{}) error) error {
+// UnmarshalYAML implements yaml.Unmarshaller
+func (cm *ColumnMapping) UnmarshalYAML(unmarshal func(interface{}) error) error {
 	type plain ColumnMapping
-	if err := unmarshal((*plain)(this)); err != nil {
-		return err
-	}
-	return nil
+	return unmarshal((*plain)(cm))
 }
 
-// Groups metric maps under a shared set of labels
+// MetricMapNamespace groups metric maps under a shared set of labels.
 type MetricMapNamespace struct {
 	labels         []string             // Label names for this namespace
 	columnMappings map[string]MetricMap // Column mappings in this namespace
 }
 
-// Stores the prometheus metric description which a given column will be mapped
-// to by the collector
+// MetricMap stores the prometheus metric description which a given column will
+// be mapped to by the collector
 type MetricMap struct {
 	discard    bool                              // Should metric be discarded during mapping?
 	vtype      prometheus.ValueType              // Prometheus valuetype
@@ -143,7 +125,8 @@ type MetricMap struct {
 
 // TODO: revisit this with the semver system
 func dumpMaps() {
-	for name, cmap := range metricMaps {
+	// TODO: make this function part of the exporter
+	for name, cmap := range builtinMetricMaps {
 		query, ok := queryOverrides[name]
 		if !ok {
 			fmt.Println(name)
@@ -160,7 +143,7 @@ func dumpMaps() {
 	}
 }
 
-var metricMaps = map[string]map[string]ColumnMapping{
+var builtinMetricMaps = map[string]map[string]ColumnMapping{
 	"pg_stat_bgwriter": {
 		"checkpoints_timed":     {COUNTER, "Number of scheduled checkpoints that have been performed", nil, nil},
 		"checkpoints_req":       {COUNTER, "Number of requested checkpoints that have been performed", nil, nil},
@@ -221,10 +204,14 @@ var metricMaps = map[string]map[string]ColumnMapping{
 		"backend_start": {DISCARD, "with time zone	Time when this process was started, i.e., when the client connected to this WAL sender", nil, nil},
 		"backend_xmin":             {DISCARD, "The current backend's xmin horizon.", nil, nil},
 		"state":                    {LABEL, "Current WAL sender state", nil, nil},
-		"sent_location":            {DISCARD, "Last transaction log position sent on this connection", nil, nil},
-		"write_location":           {DISCARD, "Last transaction log position written to disk by this standby server", nil, nil},
-		"flush_location":           {DISCARD, "Last transaction log position flushed to disk by this standby server", nil, nil},
-		"replay_location":          {DISCARD, "Last transaction log position replayed into the database on this standby server", nil, nil},
+		"sent_location":            {DISCARD, "Last transaction log position sent on this connection", nil, semver.MustParseRange("<10.0.0")},
+		"write_location":           {DISCARD, "Last transaction log position written to disk by this standby server", nil, semver.MustParseRange("<10.0.0")},
+		"flush_location":           {DISCARD, "Last transaction log position flushed to disk by this standby server", nil, semver.MustParseRange("<10.0.0")},
+		"replay_location":          {DISCARD, "Last transaction log position replayed into the database on this standby server", nil, semver.MustParseRange("<10.0.0")},
+		"sent_lsn":                 {DISCARD, "Last transaction log position sent on this connection", nil, semver.MustParseRange(">=10.0.0")},
+		"write_lsn":                {DISCARD, "Last transaction log position written to disk by this standby server", nil, semver.MustParseRange(">=10.0.0")},
+		"flush_lsn":                {DISCARD, "Last transaction log position flushed to disk by this standby server", nil, semver.MustParseRange(">=10.0.0")},
+		"replay_lsn":               {DISCARD, "Last transaction log position replayed into the database on this standby server", nil, semver.MustParseRange(">=10.0.0")},
 		"sync_priority":            {DISCARD, "Priority of this standby server for being chosen as the synchronous standby", nil, nil},
 		"sync_state":               {DISCARD, "Synchronous state of this standby server", nil, nil},
 		"slot_name":                {LABEL, "A unique, cluster-wide identifier for the replication slot", nil, semver.MustParseRange(">=9.2.0")},
@@ -238,7 +225,13 @@ var metricMaps = map[string]map[string]ColumnMapping{
 		"catalog_xmin":             {DISCARD, "The oldest transaction affecting the system catalogs that this slot needs the database to retain. VACUUM cannot remove catalog tuples deleted by any later transaction", nil, nil},
 		"restart_lsn":              {DISCARD, "The address (LSN) of oldest WAL which still might be required by the consumer of this slot and thus won't be automatically removed during checkpoints", nil, nil},
 		"pg_current_xlog_location": {DISCARD, "pg_current_xlog_location", nil, nil},
-		"pg_xlog_location_diff":    {GAUGE, "Lag in bytes between master and slave", nil, semver.MustParseRange(">=9.2.0")},
+		"pg_current_wal_lsn":       {DISCARD, "pg_current_xlog_location", nil, semver.MustParseRange(">=10.0.0")},
+		"pg_xlog_location_diff":    {GAUGE, "Lag in bytes between master and slave", nil, semver.MustParseRange(">=9.2.0 <10.0.0")},
+		"pg_wal_lsn_diff":          {GAUGE, "Lag in bytes between master and slave", nil, semver.MustParseRange(">=10.0.0")},
+		"confirmed_flush_lsn":      {DISCARD, "LSN position a consumer of a slot has confirmed flushing the data received", nil, nil},
+		"write_lag":                {DISCARD, "Time elapsed between flushing recent WAL locally and receiving notification that this standby server has written it (but not yet flushed it or applied it). This can be used to gauge the delay that synchronous_commit level remote_write incurred while committing if this server was configured as a synchronous standby.", nil, semver.MustParseRange(">=10.0.0")},
+		"flush_lag":                {DISCARD, "Time elapsed between flushing recent WAL locally and receiving notification that this standby server has written and flushed it (but not yet applied it). This can be used to gauge the delay that synchronous_commit level remote_flush incurred while committing if this server was configured as a synchronous standby.", nil, semver.MustParseRange(">=10.0.0")},
+		"replay_lag":               {DISCARD, "Time elapsed between flushing recent WAL locally and receiving notification that this standby server has written, flushed and applied it. This can be used to gauge the delay that synchronous_commit level remote_apply incurred while committing if this server was configured as a synchronous standby.", nil, semver.MustParseRange(">=10.0.0")},
 	},
 	"pg_stat_activity": {
 		"datname":         {LABEL, "Name of this database", nil, nil},
@@ -248,7 +241,7 @@ var metricMaps = map[string]map[string]ColumnMapping{
 	},
 }
 
-// Override querys are run in-place of simple namespace look ups, and provide
+// OverrideQuery 's are run in-place of simple namespace look ups, and provide
 // advanced functionality. But they have a tendency to postgres version specific.
 // There aren't too many versions, so we simply store customized versions using
 // the semver matching we do for columns.
@@ -286,11 +279,20 @@ var queryOverrides = map[string][]OverrideQuery{
 
 	"pg_stat_replication": {
 		{
-			semver.MustParseRange(">=9.2.0"),
+			semver.MustParseRange(">=10.0.0"),
 			`
 			SELECT *,
-				pg_current_xlog_location(),
-				pg_xlog_location_diff(pg_current_xlog_location(), replay_location)::float
+				(case pg_is_in_recovery() when 't' then null else pg_current_wal_lsn() end) AS pg_current_wal_lsn,
+				(case pg_is_in_recovery() when 't' then null else pg_wal_lsn_diff(pg_current_wal_lsn(), replay_lsn)::float end) AS pg_wal_lsn_diff
+			FROM pg_stat_replication
+			`,
+		},
+		{
+			semver.MustParseRange(">=9.2.0 <10.0.0"),
+			`
+			SELECT *,
+				(case pg_is_in_recovery() when 't' then null else pg_current_xlog_location() end) AS pg_current_xlog_location,
+				(case pg_is_in_recovery() when 't' then null else pg_xlog_location_diff(pg_current_xlog_location(), replay_location)::float end) AS pg_xlog_location_diff
 			FROM pg_stat_replication
 			`,
 		},
@@ -298,7 +300,7 @@ var queryOverrides = map[string][]OverrideQuery{
 			semver.MustParseRange("<9.2.0"),
 			`
 			SELECT *,
-				pg_current_xlog_location()
+				(case pg_is_in_recovery() when 't' then null else pg_current_xlog_location() end) AS pg_current_xlog_location
 			FROM pg_stat_replication
 			`,
 		},
@@ -362,24 +364,19 @@ func makeQueryOverrideMap(pgVersion semver.Version, queryOverrides map[string][]
 	return resultMap
 }
 
-// Add queries to the metricMaps and queryOverrides maps. Added queries do not
+// Add queries to the builtinMetricMaps and queryOverrides maps. Added queries do not
 // respect version requirements, because it is assumed that the user knows
 // what they are doing with their version of postgres.
 //
 // This function modifies metricMap and queryOverrideMap to contain the new
 // queries.
-// TODO: test code for all this.
+// TODO: test code for all cu.
 // TODO: use proper struct type system
 // TODO: the YAML this supports is "non-standard" - we should move away from it.
-func addQueries(queriesPath string, pgVersion semver.Version, exporterMap map[string]MetricMapNamespace, queryOverrideMap map[string]string) error {
+func addQueries(content []byte, pgVersion semver.Version, exporterMap map[string]MetricMapNamespace, queryOverrideMap map[string]string) error {
 	var extra map[string]interface{}
 
-	content, err := ioutil.ReadFile(queriesPath)
-	if err != nil {
-		return err
-	}
-
-	err = yaml.Unmarshal(content, &extra)
+	err := yaml.Unmarshal(content, &extra)
 	if err != nil {
 		return err
 	}
@@ -427,7 +424,7 @@ func addQueries(queriesPath string, pgVersion semver.Version, exporterMap map[st
 							}
 						}
 
-						// TODO: we should support this
+						// TODO: we should support cu
 						columnMapping.mapping = nil
 						// Should we support this for users?
 						columnMapping.supportedVersions = nil
@@ -622,6 +619,7 @@ func dbToFloat64(t interface{}) (float64, bool) {
 		strV := string(v)
 		result, err := strconv.ParseFloat(strV, 64)
 		if err != nil {
+			log.Infoln("Could not parse []byte:", err)
 			return math.NaN(), false
 		}
 		return result, true
@@ -662,10 +660,22 @@ func dbToString(t interface{}) (string, bool) {
 
 // Exporter collects Postgres metrics. It implements prometheus.Collector.
 type Exporter struct {
-	dsn             string
-	userQueriesPath string
-	duration, error prometheus.Gauge
-	totalScrapes    prometheus.Counter
+	// Holds a reference to the build in column mappings. Currently this is for testing purposes
+	// only, since it just points to the global.
+	builtinMetricMaps map[string]map[string]ColumnMapping
+
+	dsn              string
+	userQueriesPath  string
+	duration         prometheus.Gauge
+	error            prometheus.Gauge
+	psqlUp           prometheus.Gauge
+	userQueriesError *prometheus.GaugeVec
+	totalScrapes     prometheus.Counter
+
+	// dbDsn is the connection string used to establish the dbConnection
+	dbDsn string
+	// dbConnection is used to allow re-using the DB connection between scrapes
+	dbConnection *sql.DB
 
 	// Last version used to calculate metric map. If mismatch on scrape,
 	// then maps are recalculated.
@@ -680,8 +690,9 @@ type Exporter struct {
 // NewExporter returns a new PostgreSQL exporter for the provided DSN.
 func NewExporter(dsn string, userQueriesPath string) *Exporter {
 	return &Exporter{
-		dsn:             dsn,
-		userQueriesPath: userQueriesPath,
+		builtinMetricMaps: builtinMetricMaps,
+		dsn:               dsn,
+		userQueriesPath:   userQueriesPath,
 		duration: prometheus.NewGauge(prometheus.GaugeOpts{
 			Namespace: namespace,
 			Subsystem: exporter,
@@ -700,6 +711,17 @@ func NewExporter(dsn string, userQueriesPath string) *Exporter {
 			Name:      "last_scrape_error",
 			Help:      "Whether the last scrape of metrics from PostgreSQL resulted in an error (1 for error, 0 for success).",
 		}),
+		psqlUp: prometheus.NewGauge(prometheus.GaugeOpts{
+			Namespace: namespace,
+			Name:      "up",
+			Help:      "Whether the last scrape of metrics from PostgreSQL was able to connect to the server (1 for yes, 0 for no).",
+		}),
+		userQueriesError: prometheus.NewGaugeVec(prometheus.GaugeOpts{
+			Namespace: namespace,
+			Subsystem: exporter,
+			Name:      "user_queries_load_error",
+			Help:      "Whether the user queries file was loaded and parsed successfully (1 for error, 0 for success).",
+		}, []string{"filename", "hashsum"}),
 		metricMap:      nil,
 		queryOverrides: nil,
 	}
@@ -740,6 +762,8 @@ func (e *Exporter) Collect(ch chan<- prometheus.Metric) {
 	ch <- e.duration
 	ch <- e.totalScrapes
 	ch <- e.error
+	ch <- e.psqlUp
+	e.userQueriesError.Collect(ch)
 }
 
 func newDesc(subsystem, name, help string) *prometheus.Desc {
@@ -754,24 +778,30 @@ func newDesc(subsystem, name, help string) *prometheus.Desc {
 func queryNamespaceMapping(ch chan<- prometheus.Metric, db *sql.DB, namespace string, mapping MetricMapNamespace, queryOverrides map[string]string) ([]error, error) {
 	// Check for a query override for this namespace
 	query, found := queryOverrides[namespace]
-	if !found {
-		// No query override - do a simple * search.
-		query = fmt.Sprintf("SELECT * FROM %s;", namespace)
-	}
 
-	// Was this query disabled (i.e. nothing sensible can be queried on this
+	// Was this query disabled (i.e. nothing sensible can be queried on cu
 	// version of PostgreSQL?
-	if query == "" {
+	if query == "" && found {
 		// Return success (no pertinent data)
 		return []error{}, nil
 	}
 
 	// Don't fail on a bad scrape of one metric
-	rows, err := db.Query(query)
+	var rows *sql.Rows
+	var err error
+
+	if !found {
+		// I've no idea how to avoid this properly at the moment, but this is
+		// an admin tool so you're not injecting SQL right?
+		// nolint: gas
+		rows, err = db.Query(fmt.Sprintf("SELECT * FROM %s;", namespace))
+	} else {
+		rows, err = db.Query(query)
+	}
 	if err != nil {
 		return []error{}, errors.New(fmt.Sprintln("Error running query on database: ", namespace, err))
 	}
-	defer rows.Close()
+	defer rows.Close() // nolint: errcheck
 
 	var columnNames []string
 	columnNames, err = rows.Columns()
@@ -825,7 +855,8 @@ func queryNamespaceMapping(ch chan<- prometheus.Metric, db *sql.DB, namespace st
 				ch <- prometheus.MustNewConstMetric(metricMapping.desc, metricMapping.vtype, value, labels...)
 			} else {
 				// Unknown metric. Report as untyped if scan to float64 works, else note an error too.
-				desc := prometheus.NewDesc(fmt.Sprintf("%s_%s", namespace, columnName), fmt.Sprintf("Unknown metric from %s", namespace), nil, nil)
+				metricLabel := fmt.Sprintf("%s_%s", namespace, columnName)
+				desc := prometheus.NewDesc(metricLabel, fmt.Sprintf("Unknown metric from %s", namespace), mapping.labels, nil)
 
 				// Its not an error to fail here, since the values are
 				// unexpected anyway.
@@ -834,7 +865,6 @@ func queryNamespaceMapping(ch chan<- prometheus.Metric, db *sql.DB, namespace st
 					nonfatalErrors = append(nonfatalErrors, errors.New(fmt.Sprintln("Unparseable column type - discarding: ", namespace, columnName, err)))
 					continue
 				}
-				log.Debugln(columnName, labels)
 				ch <- prometheus.MustNewConstMetric(desc, prometheus.UntypedValue, value, labels...)
 			}
 		}
@@ -874,22 +904,44 @@ func (e *Exporter) checkMapVersions(ch chan<- prometheus.Metric, db *sql.DB) err
 	var versionString string
 	err := versionRow.Scan(&versionString)
 	if err != nil {
-		return errors.New(fmt.Sprintln("Error scanning version string:", err))
+		return fmt.Errorf("Error scanning version string: %v", err)
 	}
 	semanticVersion, err := parseVersion(versionString)
+	if err != nil {
+		return fmt.Errorf("Error parsing version string: %v", err)
+	}
+	if semanticVersion.LT(lowestSupportedVersion) {
+		log.Warnln("PostgreSQL version is lower then our lowest supported version! Got", semanticVersion.String(), "minimum supported is", lowestSupportedVersion.String())
+	}
 
 	// Check if semantic version changed and recalculate maps if needed.
 	if semanticVersion.NE(e.lastMapVersion) || e.metricMap == nil {
 		log.Infoln("Semantic Version Changed:", e.lastMapVersion.String(), "->", semanticVersion.String())
 		e.mappingMtx.Lock()
 
-		e.metricMap = makeDescMap(semanticVersion, metricMaps)
+		e.metricMap = makeDescMap(semanticVersion, e.builtinMetricMaps)
 		e.queryOverrides = makeQueryOverrideMap(semanticVersion, queryOverrides)
 		e.lastMapVersion = semanticVersion
 
 		if e.userQueriesPath != "" {
-			if err := addQueries(e.userQueriesPath, semanticVersion, e.metricMap, e.queryOverrides); err != nil {
+			// Clear the metric while a reload is happening
+			e.userQueriesError.Reset()
+
+			// Calculate the hashsum of the useQueries
+			userQueriesData, err := ioutil.ReadFile(e.userQueriesPath)
+			if err != nil {
 				log.Errorln("Failed to reload user queries:", e.userQueriesPath, err)
+				e.userQueriesError.WithLabelValues(e.userQueriesPath, "").Set(1)
+			} else {
+				hashsumStr := fmt.Sprintf("%x", sha256.Sum256(userQueriesData))
+
+				if err := addQueries(userQueriesData, semanticVersion, e.metricMap, e.queryOverrides); err != nil {
+					log.Errorln("Failed to reload user queries:", e.userQueriesPath, err)
+					e.userQueriesError.WithLabelValues(e.userQueriesPath, hashsumStr).Set(1)
+				} else {
+					// Mark user queries as successfully loaded
+					e.userQueriesError.WithLabelValues(e.userQueriesPath, hashsumStr).Set(0)
+				}
 			}
 		}
 
@@ -905,22 +957,36 @@ func (e *Exporter) checkMapVersions(ch chan<- prometheus.Metric, db *sql.DB) err
 	return nil
 }
 
-func getDB(conn string) (*sql.DB, error) {
-	if db == nil {
+func (e *Exporter) getDB(conn string) (*sql.DB, error) {
+	// Has dsn changed?
+	if (e.dbConnection != nil) && (e.dsn != e.dbDsn) {
+		err := e.dbConnection.Close()
+		log.Warnln("Error while closing obsolete DB connection:", err)
+		e.dbConnection = nil
+		e.dbDsn = ""
+	}
+
+	if e.dbConnection == nil {
 		d, err := sql.Open("postgres", conn)
 		if err != nil {
+			e.psqlUp.Set(0)
 			return nil, err
 		}
 		err = d.Ping()
 		if err != nil {
+			e.psqlUp.Set(0)
 			return nil, err
 		}
+		e.psqlUp.Set(1)
+
 		d.SetMaxOpenConns(1)
 		d.SetMaxIdleConns(1)
-		db = d
+		e.dbConnection = d
+		e.dbDsn = e.dsn
+		log.Infoln("Established new database connection.")
 	}
 
-	return db, nil
+	return e.dbConnection, nil
 }
 
 func (e *Exporter) scrape(ch chan<- prometheus.Metric) {
@@ -931,11 +997,16 @@ func (e *Exporter) scrape(ch chan<- prometheus.Metric) {
 	e.error.Set(0)
 	e.totalScrapes.Inc()
 
-	db, err := getDB(e.dsn)
+	db, err := e.getDB(e.dsn)
 	if err != nil {
 		loggableDsn := "could not parse DATA_SOURCE_NAME"
-		if pDsn, pErr := url.Parse(e.dsn); pErr != nil {
-			pDsn.User = url.UserPassword(pDsn.User.Username(), "xxx")
+		// If the DSN is parseable, log it with a blanked out password
+		pDsn, pErr := url.Parse(e.dsn)
+		if pErr == nil {
+			// Blank user info if not nil
+			if pDsn.User != nil {
+				pDsn.User = url.UserPassword(pDsn.User.Username(), "PASSWORD_REMOVED")
+			}
 			loggableDsn = pDsn.String()
 		}
 		log.Infof("Error opening connection to database (%s): %s", loggableDsn, err)
@@ -963,30 +1034,83 @@ func (e *Exporter) scrape(ch chan<- prometheus.Metric) {
 	}
 }
 
+// try to get the DataSource
+// DATA_SOURCE_NAME always wins so we do not break older versions
+// reading secrets from files wins over secrets in environment variables
+// DATA_SOURCE_NAME > DATA_SOURCE_{USER|FILE}_FILE > DATA_SOURCE_{USER|FILE}
+func getDataSource() string {
+	var dsn = os.Getenv("DATA_SOURCE_NAME")
+	if len(dsn) == 0 {
+		var user string
+		var pass string
+
+		if len(os.Getenv("DATA_SOURCE_USER_FILE")) != 0 {
+			fileContents, err := ioutil.ReadFile(os.Getenv("DATA_SOURCE_USER_FILE"))
+			if err != nil {
+				panic(err)
+			}
+			user = strings.TrimSpace(string(fileContents))
+		} else {
+			user = os.Getenv("DATA_SOURCE_USER")
+		}
+
+		if len(os.Getenv("DATA_SOURCE_PASS_FILE")) != 0 {
+			fileContents, err := ioutil.ReadFile(os.Getenv("DATA_SOURCE_PASS_FILE"))
+			if err != nil {
+				panic(err)
+			}
+			pass = strings.TrimSpace(string(fileContents))
+		} else {
+			pass = os.Getenv("DATA_SOURCE_PASS")
+		}
+
+		uri := os.Getenv("DATA_SOURCE_URI")
+		dsn = "postgresql://" + user + ":" + pass + "@" + uri
+	}
+
+	return dsn
+}
+
 func main() {
-	flag.Parse()
+	kingpin.Version(fmt.Sprintf("postgres_exporter %s (built with %s)\n", Version, runtime.Version()))
+	log.AddFlags(kingpin.CommandLine)
+	kingpin.Parse()
+
+	// landingPage contains the HTML served at '/'.
+	// TODO: Make this nicer and more informative.
+	var landingPage = []byte(`<html>
+	<head><title>Postgres exporter</title></head>
+	<body>
+	<h1>Postgres exporter</h1>
+	<p><a href='` + *metricPath + `'>Metrics</a></p>
+	</body>
+	</html>
+	`)
 
 	if *onlyDumpMaps {
 		dumpMaps()
 		return
 	}
 
-	dsn := os.Getenv("DATA_SOURCE_NAME")
+	dsn := getDataSource()
 	if len(dsn) == 0 {
-		log.Fatal("couldn't find environment variable DATA_SOURCE_NAME")
+		log.Fatal("couldn't find environment variables describing the datasource to use")
 	}
 
 	exporter := NewExporter(dsn, *queriesPath)
+	defer func() {
+		if exporter.dbConnection != nil {
+			exporter.dbConnection.Close() // nolint: errcheck
+		}
+	}()
+
 	prometheus.MustRegister(exporter)
 
-	http.Handle(*metricPath, prometheus.Handler())
+	http.Handle(*metricPath, promhttp.Handler())
 	http.HandleFunc("/", func(w http.ResponseWriter, r *http.Request) {
-		w.Write(landingPage)
+		w.Write(landingPage) // nolint: errcheck
 	})
 
 	log.Infof("Starting Server: %s", *listenAddress)
 	log.Fatal(http.ListenAndServe(*listenAddress, nil))
-	if db != nil {
-		defer db.Close()
-	}
 }
